#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Aug 16 10:45:38 2017

@author: IsabellTrinh
"""

from UnitKeeper import UnitKeeper as uk
from random import randrange, shuffle
from FlaskData import FlaskData

import numpy as np
import scipy.io as sio


'''
The Flask class is responsible for keeping track of/considering physics and logic, initiates units/chains etc.
'''

class Flask(FlaskData):
    uk = None 
    
    def __init__(self, size, binsize, logging):
    #initialize Unitkeeper in Flask
        self.uk = uk(size)
        self.binsize = binsize
        self.logging = logging
    
    
    '''
    CHAIN GENERATION
    '''
    
    #TODO use disorder/order spectra for making chains?            
    def makeChain(self, length, props, lts):
        '''
        Creates a chain of the given length at a random startposition in a number of attempts a.
        '''
        a = 100 #give number of wanted attempts to make a new chain
        i = 0
        
        while i != a:
            #define startpos to generate a random starting position
            startpos = ()
            
            #choose start position depending on the dimension and boundaries: 
            #get a random number between the boundaries for each dimension
            for i1 in range(0, self.uk.dim): #choose start position depending on 
                startpos = startpos + (randrange(self.uk.boundaries[i1, 0], self.uk.boundaries[i1, 1]),)
            
            #find a new start position if the one generated in the for-loop is already occupied
            if self.uk.isOccupied(startpos):
                i = i + 1
                #self.logging.debug("Position "+str(startpos)+" is already occupied.")
            
            #is the start position is empty, try to find a path from this position 
            else:
                newChain = self._findPath(startpos, [], length)
             
                #if there is no space for the chain, find a new start position
                if not newChain:
                    i = i + 1
                    self.logging.debug("No Chain could be generated at position "+str(startpos)+".")
                
                #chain could be generated --> add the chain to the lattice and end loop.
                else:
                    self.uk.addChain(poss=newChain, props=props, lts=lts)
                    self.logging.debug("Chain found in %.f attempts."%(i+1))
                    i = a
        
        if not newChain:
            self.logging.debug("No suitable chain could be found in %.f attempts."%i) 
            
            
            
    def _findPath(self, pos, vpos, length): #give pos as array, position has to be unoccupied
        '''
        Identifies a possible chain of a given length from a startposition through 
        recursive backtracking.
        Give "pos" as tuple, position has to be unoccupied, give vpos as [], it 
        gives already visited lattice sites to the function in the recursion.
        '''       
        #append the current position to already visited positions.
        vpos.append(pos)
        
        #chain has been generated (length is subtracted until it reaches 1, because the startposition is already given.)
        if length == 1:
            return [pos]
        
        #get list of possible moves
        ols = self.uk.getOpenLatticeSites(pos)
        
        #remove all previously visited positions during the findPath call (vpos) 
        ols = [x for x in ols if x not in vpos]
        
        #randomize list so that the chain shape is random
        shuffle(ols)
        
        #if there are no open lattice sites left, but the chain hasn't reached its length, backtrack.
        if not ols:
            return None
        
        #recursion over for-loop 
        else:
            #first unit in list is used in the next findPath instance
            for u in ols:
                nextpos = self._findPath(u, vpos, length-1)
                #if there is a next position available save the position in the path.
                if nextpos != None:
                    return [pos] + nextpos


    '''
    PROTEIN GENERATION
    '''
    
    def getProteinPositions(self, posname):
        '''
        Generates positions for a protein from a matlabfile (array). Input the file by 
        its filename (case-sensitive) as a string.
        '''
        #generate positions array
        poss = np.zeros([0, self.uk.dim], dtype = int)
        
        #import matlab file and get array. Value is '1' if a protein unit is there.
        mat_contents = sio.loadmat(posname + '.mat')
        protein = mat_contents[posname]
        
        #Case 2D: generate positions
        if self.uk.dim == 2:
            #get all positions where the value is 1 and put it into positions array.
            i, j = np.where(protein == 1)
            for x in range(len(i)):
                poss = np.append(poss, [[i[x], j[x]]], 0)
        
        #Case 3D: generate positions
        if self.uk.dim == 3:
            #get all positions where the value is 1 and put it into positions array.
            i, j, k = np.where(protein == 1)
            for x in range(len(i)):
                poss = np.append(poss, [[i[x], j[x], k[x]]], 0)
        
        if len(poss) != 0:        
            return poss
        
        else:
            self.logging.error("Error_inputProteinPositions: No protein could be found in the file.")
            return None
        
            
    def getProteinProperties(self, poss, hydroname, chargename):
        '''
        Gets the properties (hydrophobicity and charge) from array matlab files. 
        Input the file by its filename (without .mat and case-sensitive) as a string
        '''
        #import matlab file and get array for hydrophobicity and charges.
        mat_contents = sio.loadmat(hydroname + '.mat')
        hydrophobicity = mat_contents[hydroname]
        mat_contents = sio.loadmat(chargename + '.mat')
        charges = mat_contents[chargename]
        
        #get values at positions of the protein.
        h = [hydrophobicity[tuple(x)] for x in poss]
        c = [charges[tuple(x)] for x in poss]
        
        props = np.column_stack((h, c))
        
        return props
    
    
    def makeProtein(self, startpos, posname, hydroname, chargename):
        poss = self.getProteinPositions(posname)
        props = self.getProteinProperties(poss, hydroname, chargename)
        
        #get relative positions from one entry to the next
        diff = np.diff(poss, axis = 0)
        
        #get final positions with startposition
        #TODO change if clause
        diff = np.insert(diff, 0, startpos, axis=0)
        finalposs = np.cumsum(diff, axis = 0) 
        
        if (finalposs > -1).all() and (finalposs < self.uk.boundaries[0][1]).all():
            self.uk.addProtein(poss=finalposs, props=props)  
            return finalposs, props
        else:
            self.logging.error("Start position not possible.")
      
        

    '''
    MOVEMENT
    '''
    
    def canMove(self, ident):
        '''
        Returns a boolean, 'true' if there is an open lattice spot the unit can 
        move to, 'false' if there is none.
        '''
        if self.uk.isValidIdent(ident):
            #get list of possible positions to move to for given unit.
            pm = self.getPossibleMovement(ident)
            
            #returns 'true' if list is not empty, 'false' if list is empty.
            return pm != []
        
        else:
            self.logging.error("Error_canMove: ident does not exist.")

    def getPossibleMovement(self, ident):
        '''
        Gets the positions the given unit can move to and returns empty list/array 
        if there are none.
        '''
        if self.uk.isValidIdent(ident):
            
            #get chain neighbour(s)
            cn = self.uk.getChainNeighbours(ident)
            
            
            #Case: Given Unit is at the end of a Chain
            if self.uk.isChainEnd(ident):
                
                #get open lattice sites of the given unit, transform into list 
                #so that it's possible to use "choice"
                pposs = self.uk.getOpenLatticeSites(self.uk.getPos(cn[0]))
                
                #return list of possible sites
                return pposs
                
            #Case: Given Unit is in the middle of a Chain
            else:
                #get the open lattice sites of the given unit's neighbours
                a = self.uk.getOpenLatticeSites(self.uk.getPos(cn[0]))
                b = self.uk.getOpenLatticeSites(self.uk.getPos(cn[1]))
                
                #convert arrays a and b to sets to get intersection
                aset = set(a)
                bset = set(b)
                
                #get possible positions to move to by getting the overlap between 
                #the neighbour's neighbours
                pposs = set(aset).intersection(bset)
                
                #Convert to tuple, because "choice" does not take sets.
                return list(pposs)  
    
    def move(self, ids, newPoss):
        self.uk.moveChainPart(ids, newPoss)
        
    
    '''SINGLE UNIT MOVEMENTS'''
    def singleMove(self, ident):
        '''
        Moves Unit to a new position depending on its linktype and updates its
        ident and position in all arrays.
        '''
        #TODO: implement for rigid neighbours
        
        #Case: Unit is connected to its neighbours flexibly.
        if not self.uk.getRigidNeighbours(ident):
            self.flexibleMove(ident)
        
        #Case: Unit is connected to one or two of its neighbours rigidly.
        else:
            raise NotImplementedError("Rigid movement is not implemented yet.")


    def flexibleMove(self, ident):
        '''
        Returns a list of positions the unit can be moved to, if the unit is
        connected to its neighbours by a flexible link.
        '''
        
        if self.uk.isValidIdent(ident):
            oldpos = self.uk.getPos(ident)
            if self.canMove(ident):
                #get possible positions the given unit can move to, if the unit can move.
                pposs = self.getPossibleMovement(ident)
                
                return [ident], [oldpos], pposs
            else:
                return None
                self.logging.error("Unit cannot be moved.")     
                
        else:
            self.logging.error("Error_moveUnit: Given Unit does not exist or new position"
                  + "is not within bounds/valid.")        


    '''MULTIPLE UNIT MOVEMENTS'''
                 
    '''ROTATION'''    
    def rotate(self, startIdent, endIdent, angle):
        '''
        Give the start and end of the chain part to be rotated.
        '''
        #Case1: startIdent is the end of a chain and endIdent is somewhere in the middle of the chain
        if self.uk.isChainEnd(startIdent) and not self.uk.isChainEnd(endIdent):
            if startIdent < endIdent:
                return self._rotateEnd(endIdent+1, startIdent, angle)
                self.logging.debug("Case1.1: startIdent is at the start of a chain.")
            else:
                return self._rotateEnd(endIdent-1, startIdent, angle)
                self.logging.debug("Case1.2: startIdent is at the end of a chain.")

                
        #Case2: endIdent is the end of a chain and startIdent is somewhere in the middle of the chain
        elif self.uk.isChainEnd(endIdent) and not self.uk.isChainEnd(startIdent):
            if endIdent < startIdent:
                return self._rotateEnd(startIdent+1, endIdent, angle)
                self.logging.debug("Case2.1: endIdent is at the start of a chain.")

            else:
                return self._rotateEnd(startIdent-1, endIdent, angle)
                self.logging.debug("Case2.2: endIdent is at the end of a chain.")
                
        #Case3: If start and endIdent are both at the end of a chain.
        elif self.uk.isChainEnd(startIdent) and self.uk.isChainEnd(endIdent):
            return self._rotateEnd(startIdent, endIdent, angle)
        
        #Case4: both start and endIdent are in the middle of the chain.
        else:
            return self._rotateMiddle(startIdent, endIdent, angle)
            
                    
    
    def _rotateEnd(self, originIdent, endIdent, angle):
        '''
        Returns a list of positions to rotate part of the chain around origin unit. 
        EndIdent is the end to the region to be rotated.
        Positive angles rotate counter clockwise and negative angles rotate clockwise.
        ''' 
        #initiate chainIdent and first unit in chain
        chainIdent = self.uk.getChainIdent(originIdent)
        first = self.uk.chains[chainIdent][0]
        
        #IDs to move:
        ids = range(min(originIdent, endIdent), max(originIdent, endIdent)+1)
        self.logging.debug("ids: " + str(ids))
        
        #initiate sin and cos
        c, s = np.cos(np.radians(angle)), np.sin(np.radians(angle))
        
        #get position of origin
        opos = self.uk.getPos(originIdent)
        self.logging.debug("original position:"+str(opos))
        self.logging.debug("angle: %.f"% angle )
        
        #get list of old positions
        oldposs = [tuple(self.uk.ids[x]) for x in ids]
        
        #if the list passes through periodic boundary 
        if self.uk.isInPeriodicBoundary(chainIdent):
            #get list of positions corrected with periodic boundaries reversed
            poss = self.uk.reversePeriodicBoundary(chainIdent)[1]
            #correct origin position to one with reversed periodic boundaries
            opos = poss[originIdent-first]
            self.logging.debug("corrected original position:"+str(opos))
            #generate list of positions to rotate
            rotateposs = poss[ids[0]-first:ids[-1]-first+1]
        else:
            #generate list of positions to rotate
            rotateposs = [self.uk.getPos(x) for x in ids]

            
        #subtract origin from positions to rotate
        sub = [np.subtract(pos, opos) for pos in rotateposs]
        
        
        #Case: 2D
        if self.uk.dim == 2:
            #initiate rotation matrix with angle
            R = np.matrix([[c, -s], [s, c]], dtype = int)
            
            #rotate and turn into array after dot product (.A1)
            rot = [np.dot(R, x).A1 for x in sub]
            
            #add origin to positions to get final positions
            temp = [tuple(np.add(pos, opos)) for pos in rot]
            #apply boundary conditions
            mod = [self.uk.applyPeriodicBoundary(pos) for pos in temp]
            newposs = [pos for pos in mod if self.uk.isValidPos(pos) and (pos in oldposs or not self.uk.isOccupied(pos))]
            
            #check if move is possible
            if temp == newposs:
                return ids, oldposs, newposs 

            else:
                self.logging.debug("Rotation was not possible.")
                return None
        
         
        #Case: 3D
        elif self.uk.dim == 3:
            #initiate rotation matrices for different axes
            R_x = np.matrix([[1, 0, 0], [0, c, -s], [0, s, c]], dtype = int)
            R_y = np.matrix([[c, 0, s], [0, 1, 0], [-s, 0, c]], dtype = int)
            R_z = np.matrix([[c, -s, 0], [s, c, 0], [0, 0, 1]], dtype = int)
            R_list = [R_x, R_y, R_z]
            
            for R in R_list:
                #rotate and turn into array after dot product (.A1)
                rot = [np.dot(R, pos).A1 for pos in sub]
                
                #add origin to positions to get final positions
                temp = [tuple(np.add(pos, opos)) for pos in rot]
                
                #apply boundary conditions
                mod = [self.uk.applyPeriodicBoundary(pos) for pos in temp]
                
                #check if it's possible to move to positions.
                newposs = [pos for pos in mod if self.uk.isValidPos(pos) and (pos in oldposs or not self.uk.isOccupied(pos))]
                
                if mod == newposs:
                    return ids, oldposs, newposs 
                
            if not mod == newposs:
                self.logging.debug("Rotation was not possible.")
                return None
                
        else:
            self.logging.debug("Error_rotateEnd: Cannot rotate in more than 3 dimensions.")

            
    def _rotateMiddle(startIdent, endIdent):
        
        pass

    
    '''TRANSLATION'''
    def translateChain(self, chainIdent, shift):
        '''
        Returns a list of positions to move the chain by a certain number of lattice rows/columns. Shift should have the form of an array corresponding to the dimensions (x, y, z)]
        '''
        
        #check if chain ident is valid and if the shift has the right dimension
        if chainIdent <= self.uk.links[-1][0] and len(shift) == self.uk.dim:
            #get list of idents in the chain
            chain = self.uk.chains[chainIdent]
            
            #get list of old positions
            oldposs = [self.uk.getPos(ident) for ident in chain]
            
            #calculate new positions
            temp = [tuple(np.add(pos, shift).tolist()) for pos in oldposs]
            
            #apply boundary condition
            mod = [self.uk.applyPeriodicBoundary(pos) for pos in temp]
            
            #check if the new positions can be moved to.
            newposs = [pos for pos in mod if self.uk.isValidPos(pos) and (not self.uk.isOccupied(pos) or pos in oldposs)]
            
            #if the position for the chain move is open, move the units
            if newposs == mod:
                return chain, oldposs, newposs
            else: 
                self.logging.debug("Translation was not possible.")
                return None
        else:
            self.logging.debug("Error_translateChain: The given chain does not exist or the shift does not have the right dimensionality.")
      
    
    '''SNAKE'''
    def snake(self, head, headPos):
        '''
        Chain moves behind start or end unit by one lattice site. "head" should be either at the end or start of a chain. Makes chain able to move in crowded environments.
        '''
        #TODO adjust for rigid movement
        if self.uk.isValidIdent(head):
            
            #get list of idents in the chain
            chainIdent = self.uk.links[head][0]
            chain = self.uk.chains[chainIdent]
            
            if self.uk.isValidPos(headPos) and not self.uk.isOccupied(headPos):    
                #make list of current positions of the chain
                poss = [self.uk.getPos(x) for x in chain]
                oldposs = list(poss)

                #Case: when the head is at the start of the chain (ident wise)
                if head == chain[0]:
                    #generate list of new positions for the chain
                    poss.insert(0, headPos)
                    poss.pop()
                    
                    #apply boundary conditions
                    newposs = [self.uk.applyPeriodicBoundary(pos) for pos in poss]
                    
                    return chain, oldposs, newposs
                
                #Case: when the head is at the end of the chain (ident wise)
                elif head == chain[-1]:
                    poss.append(headPos)
                    poss.pop(0)
                    
                    #apply boundary conditions
                    newposs = [self.uk.applyPeriodicBoundary(pos) for pos in poss]
                    
                    return chain, oldposs, poss
            
                else:
                    self.logging.debug("Given chain head is not a chain head.")
                    return None
            
            else:
                self.logging.debug("New head position is not valid or already occupied.")
                return None
            
        else:
            self.logging.error("Error_snake: Given chain does not exist.")
            return None
      

    '''
    AMINO ACID AVERAGING
    '''
    def getProteinAverage(self, sequence):
        '''
        Average hydrophobicity and charge of the amino acids in a protein over a number of residues (binsize).
        '''
        #initizalize hydrophobic scale and charges
        #hscales = {'F' : 100, 'I' : 99, 'W' : 97, 'L' : 97, 'V' : 76, 'M' : 74, 'Y' : 63,' C' : 49, 'A' : 41, 'T' : 13, 'H' : 8, 'G' : 0, 'S' : -5, 'Q' : -10, 'R' : -14, 'K' : -23, 'N' : -28, 'E' : -31, 'P' : -46, 'D' : -55}
        #charges = {'F' : 0, 'I' : 0, 'W' : 0, 'L' : 0, 'V' : 0, 'M' : 0, 'Y' : 0,' C' : 0, 'A' : 0, 'T' : 0, 'H' : 1, 'G' : 0, 'S' : 0, 'Q' : 0, 'R' : 1, 'K' : 1, 'N' : 0, 'E' : -1, 'P' : 0, 'D' : -1}
        
        charges = {'A' : 6.0, 'C' : 5.1, 'D' : 2.8, 'E' : 3.2, 'F' : 5.5, 'G' : 6.0, 'H' : 7.6, 'I' : 5.9,'K' : 9.7, 'L' : 6.0, 'M' : 5.7, 'N' : 5.4, 'P' : 6.5, 'Q' : 5.7, 'R' : 10.8, 'S' : 5.7, 'T' : 5.9, 'V' : 6.0, 'W' : 6.0, 'Y' : 5.7}
        hscales = {'A' : 1.0, 'C' : 0.2, 'D' : -3, 'E' : -2.6, 'F' : 2.5, 'G' : 0.7, 'H' : -1.7, 'I' : 3.1, 'K' : -4.6, 'L' : 2.2, 'M' : 1.1, 'N' : -2.7,'P' : -0.3, 'Q' : -2.9, 'R' : -7.5, 'S' : -1.1, 'T' : -0.8, 'V' : 2.3, 'W' : 1.5, 'Y' : 0.1}
        
        #determine number of units
        un = len(sequence)/self.binsize
        i = 0
        props = np.zeros([0,2], dtype = int)

        #run while loop for the number of bins needed for the protein.
        while i < un:
            #initialize hydrophobicity and charge variables
            h = 0
            c = 0
            
            start = self.binsize*i
            
            #calculate
            for letter in sequence[start:start+(self.binsize)]:
                #calculate the hydrophobicity and charge of the amino acids in the bin
                h = h + hscales[letter]
                c = c + charges[letter]
            props = np.append(props, [[h/self.binsize, c/self.binsize]], 0)
            
            i += 1
        
        return props    

    
        
    '''
    ENERGY
    ''' 
    def _freeEnergyPos(self, pos, prop):
        '''
        Calculates free energy value for a given position.
        '''
        E_u = 0
        
        #get list of neighbouring sites
        N = self.uk.getNeighboursInBoundary(pos)
        
        #filter for occupied sites
        unitN = [x for x in N if self.uk.isOccupied(x)]
        
        #get list of empty neighbouring sites - NOT CONSIDERED
        #emptyN = [x for x in N if x not in unitN]
        
        #get properties of neighbouring units if there are any.
        if unitN:
            #get the properties of the neighbouring units
            propsN = [self.uk.getProp(self.uk.getIdent(x)) for x in unitN]
            for pn in propsN:
                E_HCI, E_CCI = self.CItoEnergy(props1=prop, props2=pn)
                E_u += E_HCI + E_CCI
        return E_u, len(propsN)
    
    def _compatibilityIndices(self, props1, props2):
        '''
        Calculates hydrophobic compatibility index for a given position.
        '''
        #initiate hydrophobic and charge compatibility indices
        HCI = 0
        CCI = 0 
        
        HCI = HCI + (20 - np.absolute((props1[0] - props2[0])*(19/10.6)))
        CCI = CCI + (11 - ((props1[1]-7) * (props2[1]-7)*(19/33.8)))
            
        return HCI, CCI

    def CItoEnergy(self, props1, props2):
        #get compatibility indices
        CIs = self._compatibilityIndices(props1, props2)
        
        #subtract 20, to get numbers ranging from -10 to +10 and multiply by factor
        #to get the right relation between charge and hydrophobic free energies
        #hydrophobic interactions (burial of one CH2-group) in the range of 0.6-1.6 kcal/mol
        #electrostatic interactions in the range of 1.4 kcal/mol
        E_hyd = 0.9 #kcal/mol
        E_ele = 1.4 #kcal/mol
        
        E_HCI = (10-CIs[0])*E_hyd
        E_CCI = (10-CIs[1])*E_ele
        
        return E_HCI, E_CCI
    
    
    def freeEnergyUnit(self, ident):
        '''
        Calculates free energy value for one unit.
        '''
        #get position, properties and neighbours of ident
        pos = self.uk.getPos(ident)
        prop = self.uk.getProp(ident)
        
        return self._freeEnergyPos(pos, prop)
            
   
    def freeEnergyChain(self, chainIdent):
        '''
        Calculates a free energy value of one chain.
        '''
        E_c = 0
        connN = 0
        #get list of idents in the chain
        chain = self.uk.chains[chainIdent]
        
        #Get energy for each unit in the chain
        for unit in chain:
            dE, dN = self.freeEnergyUnit(unit)
            E_c += dE
            connN += int(dN)
        
        return E_c, connN
        
           
    def freeEnergyDiff(self, ids, oldposs, newposs):
        '''
        calculates the difference in free energies between the new and old position 
        of one or more units.
        '''
        E_new = 0
        E_old = 0
        N_old = 0
        N_new = 0
        
        #calculate energy for old positions
        for i, unit in enumerate(oldposs):
            id1 = ids[i]
            E_u = 0
            #get list of neighbouring sites
            N = self.uk.getNeighboursInBoundary(unit)
            
            #filter for occupied sites
            unitN = [x for x in N if self.uk.isOccupied(x)]
            N_old += len(unitN)
            prop = self.uk.getProp(id1)
            if unitN:
                #get the properties of the neighbouring units
                for x in unitN:
                    pn = self.uk.getProp(self.uk.getIdent(x))
                    E_HCI, E_CCI = self.CItoEnergy(props1=prop, props2=pn)
                    id2 = self.uk.getIdent(x)
                    if abs(id2-id1) == 1 or not(id2 in ids):
                        E_u += E_HCI + E_CCI
                    else:
                        E_u += (E_HCI + E_CCI)/2
                        #print(id1, id2, 'old')
            
            E_old += E_u
            
        
        #calculate energy for new positions
        for i, pos in enumerate(newposs):
            id1 = ids[i]
            
            #initialize energy for unit
            E_u = 0
            
            #get property of position
            prop = self.uk.getProp(self.uk.getIdent(oldposs[i]))            

            #get list of neighbouring sites
            n = self.uk.getNeighboursInBoundary(pos)
            
            #filter for occupied sites and remove positions that are occupied
            #by old chain positions
            cur_n = [x for x in n if self.uk.isOccupied(x) and not (x in oldposs)]
            
            #filter neighbouring sites for sites occupied by the new chain
            new_n = [x for x in n if x in newposs]
            
            #make sure that there are no doubles in the neighbouring sites
            if set(new_n).intersection(cur_n):
                self.logging.error("Error_freeEnergyDiff: neighbours present in both current and new lists.")
            
            #calculate interaction energy with current neighbours if there are any
            if cur_n:
                #get the properties of the neighbouring units
                cur_props = [self.uk.getProp(self.uk.getIdent(x)) for x in cur_n]
                
                #calculate energies
                for cur_prop in cur_props:
                    E_HCI, E_CCI = self.CItoEnergy(props1=prop, props2=cur_prop)
                    E_u += E_HCI + E_CCI
                    N_new += 1
            
            #calculate interaction energy neighbours of new chain.
            if new_n:
                #get the properties of chain sites
                new_ids = [ids[newposs.index(x)] for x in new_n]
                new_props = [self.uk.getProp(ids[newposs.index(x)]) for x in new_n]
                
                N_new += len(new_props)
                
                #calculate energies
                for j, new_prop in enumerate(new_props):
                    E_HCI, E_CCI = self.CItoEnergy(props1 = prop, props2 = new_prop)
                    id2 = new_ids[j]
                    if abs(id2-id1) == 1:
                        E_u += E_HCI + E_CCI
                    else:
                        E_u += (E_HCI + E_CCI)/2
                        #print(id1, id2, 'new')
                    
            E_new += E_u
         
        #return difference, if the value is negative, then the new confirmation is better
        return 2*(E_new - E_old), 2*(N_new - N_old)
            
    def getNeighboursCas13a(self):
        number_n = 0
        poss = [self.uk.getPos(unit) for unit in self.uk.chains[0]]
        
        for pos in poss:
            #get list of neighbouring sites
            n = self.uk.getNeighboursInBoundary(pos)
            
            #filter for occupied sites and remove positions that are occupied
            #by same chain
            realN = [x for x in n if self.uk.isOccupied(x) and not (x in poss)]
            
            #count neighbours
            number_n += len(realN)
            
        return number_n
    
    def getProteinDistanceFromCas13a(self):
        #get position of first unit of Cas13a
        CasPos = self.uk.getPos(self.uk.chains[0][0])
        
        #get positions of all first units of chains in lattice
        chainPoss = [self.uk.getPos(self.uk.chains[x][0]) for x in range(1, len(self.uk.chains))]
        
        #calculate distances for between Cas13a and all other chains.
        distances = [((x[0]-CasPos[0])**2+(x[1]-CasPos[1])**2+(x[2]-CasPos[2])**2)**(1/2) for x in chainPoss]
        
        return distances
        
    
    
    def getFreeEnergy(self):
        Energy = 0
        N=0
        
        for i in range(0, len(self.uk.chains)):
            dE, dN = self.freeEnergyChain(i)
            Energy += dE
            N += int(dN)
        
        return Energy, N
    

        
        
    
    
    '''
    VISUALIZATION

         
    def draw(self, tries, concentration):
        if self.uk.dim == 1:
            print("1D plotting not supported yet.")
        elif self.uk.dim == 2:
            self._draw2D(tries, concentration)
        elif self.uk.dim == 3:
            self._draw3D(tries, concentration)
        else:
            print("Ufff, more than 3D? How am I supposed to do that?")
            
    def _draw2D(self, tries, concentration):
        #scatterplot 
        
        #check dimension
        if self.uk.dim == 2:
            
            fig = plt.figure()
            
            #set boundaries for plot
            plt.xlim(-1, self.uk.boundaries[1][1]+1)
            plt.ylim(self.uk.boundaries[0][1]+1, -1)
            
            #get all positions of water/empty sites and plot as scatterplot
            a, b = np.argwhere(self.uk.poss == -1).T
            plt.scatter(b, a, facecolors='none', edgecolors='grey')
            
            #plot chains: for all chains in chain list
            for chain in self.uk.chains:
                chainIdent = self.uk.chains.index(chain)
                if self.uk.isInPeriodicBoundary(chainIdent):
                    sections = self.uk.getUnitsInPeriodicBoundary(chainIdent)
                    [self._plotChain(section, 'C1') for section in sections]
                else:
                    self._plotChain(chain, 'lightblue')                
            return fig
        else:
            print("Error_draw2D: Given lattice is not 2 dimensional.")
   
         
    def _draw3D(self, tries, concentration):
        #scatterplot
        if self.uk.dim == 3:
            fig = plt.figure()
            ax = fig.add_subplot(111, projection='3d')
            
            #set limits for x, y, z axes
            ax.set_xlim(self.uk.boundaries[0])
            ax.set_ylim(self.uk.boundaries[1])
            ax.set_zlim(self.uk.boundaries[2])

            #plot Cas13a
            chain = self.uk.chains[0]
            chainIdent = self.uk.chains.index(chain)
            sections = self.uk.getUnitsInPeriodicBoundary(chainIdent)
            for section in sections: 
                #get start and end unitID of chain (part)
                start = section[0]
                end = section[-1]
            
                #get position of IDs in chain and plot
                x = self.uk.ids[start:end+1,0]
                y = self.uk.ids[start:end+1,1]
                z = self.uk.ids[start:end+1,2]
                plt.plot(x, z, y, '-o', color = 'lightblue')
            
            #plot the chains
            for chain in self.uk.chains[1:]:
                chainIdent = self.uk.chains.index(chain)
                if self.uk.isInPeriodicBoundary(chainIdent):
                    sections = self.uk.getUnitsInPeriodicBoundary(chainIdent)
                    [self._plotChain(section) for section in sections]
                else:
                    self._plotChain(chain)
            
            #set label ticks to integer
            ax.xaxis.set_major_locator(MaxNLocator(integer=True))
            ax.yaxis.set_major_locator(MaxNLocator(integer=True))
            ax.zaxis.set_major_locator(MaxNLocator(integer=True))
            
            #set labels for axis
            ax.set_xlabel('x')
            ax.set_ylabel('y')
            ax.set_zlabel('z')
            
            plt.show()
            fig.savefig("metropolis" + str(self.uk.size) + str(concentration) + str(tries) + ".png", transparent = True, dpi = 300)
            
            return fig
        else:
            print("Error_draw3D: Given lattice is not 3 dimensional.")
            
    def _plotChain(self, idents):
        #get start and end unitID of chain (part)
        start = idents[0]
        end = idents[-1]
        
        #get position of IDs in chain and plot
        x = self.uk.ids[start:end+1,0]
        y = self.uk.ids[start:end+1,1]
        if self.uk.dim == 2:
            plt.plot(y, x, '-o')
            
        elif self.uk.dim == 3:
            z = self.uk.ids[start:end+1,2]
            plt.plot(x, z, y, '-o')
        
    '''   