#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Oct 25 17:43:48 2017

@author: IsabellTrinh
"""
from Simulator import Simulator
import scipy.constants as spc
import numpy as np

class TDP:
    CAHS94205_2 = "CAHS94205_2"
    CAHS106094 = "CAHS106094"
    CAHS107838 = "CAHS107838"
    SAHS33020 = "SAHS33020"
    SAHS68234 = "SAHS68234"
    RANDOM_HYDROPHOB= "rand_HydroPhob"
    GFP = "GFP"
    RANDOM_HYDROPHIL = "rand_HydroPhil"
    
    _molarMasses = {
                CAHS94205_2 : 25485.32,
                CAHS106094 : 26777.02,
                CAHS107838 : 26559.87,
                SAHS33020 : 18752.46,
                SAHS68234 : 19369.12,
                RANDOM_HYDROPHOB : 26974.660,
                RANDOM_HYDROPHIL : 27386.685,
                GFP : 26886.272,
            }
    sequences = {
                CAHS94205_2 : 'MSGRNVESHMERNEKVVVNNSGHADVKKQQQQVEHTEFTHTEVKAPLIHPAPPIISTGAAGLAEEIVGQGFTASAARISGGTAEVHLQPSAAMTEEARRDQERYRQEQESIAKQQEREMEKKTEAYRKTAEAEAEKIRKELEKQHARDVEFRKDLIESTIDRQKREVDLEAKMAKRELDREGQLAKEALERSRLATNVEVNFDSAAGHTVSGGTTISSSDKMEIKRN',
                CAHS106094 : 'MEAMNMNIPRDAMFVPPPESEQNGYHEKSEVQQTSYMQSQVKVPHYNFPTPYFTTSFSAQELLGEGFQASISRISAVTEDMQSMEIPEFVEEARRDYAAKTRENEMLGQQYEKELERKSEAYRKHQEVEADKIRKELEKQHMRDIEFRKEIAELAIENQKRMIDLECRYAKKDMDRERTKVRMMLEQQKFHSDIQVNLDSSAAGTESGGHVVSQSEKFTERNREMKR',
                CAHS107838 : 'MSAEAMNMNMNQDAVFIPPPEGEQYERKEKQEIQQTSYLQSQVKVPLVNLPAPFFSTSFSAQEILGEGFQASISRISAVSEELSSIEIPELAEEARRDFAAKTREQEMLSANYQKEVERKTEAYRKQQEVEADKIRKELEKQHLRDVEFRKDIVEMAIENQKKMIDVESRYAKKDMDRERVKVRMMLEQQKFHSDIQVNLDSSAAGTETGGQVVSESQKFTERNRQIKQ',
                SAHS33020 : 'MARFLVALALFGVVAMTAASGDAPKEWSGKPWLGKFVAEVSDKSENWEAFVDALGLPDQYPRAQLKTIHSFYKQGEHYHHILSLPDKNINKDIEFTLGQEVEIKHGEHSLKIKYFEDGNKLVADVSIPAKGKSIHDVYDVQGDQLIKSYKVGDVVAKKWFKKVANPAA',
                SAHS68234 : 'MARFLVALALFGVVAMTAATGDAPKEWSGKPWLGKFVAEVTDKSENWEAFVDALGLPEQFGRAPVKTIQKIYKQGDHYHHIFALPDKNFEKDIEFTLGQEVEIKQGEHIAKTKYSEDGEKLVADVSIPTKGKTIRSEYEVQGDQLIKTYKTGDIVAKKWFKKVANPTEAPAQAA',
                RANDOM_HYDROPHOB : 'MFWIWYCAFCIYVFIIGLIGVVGLCFYYAGMMAIYAAILCFGGWCVYIAFMFVMGVIGVYWGVIMYWCFYCVWIIIAYALFVMLMMLWGGWYIAIFVIFVVWLCVCYYWGGWGYMWIAYVIMALYWVYLAMFYWAWIVVCMFAMWLVGYFLMMAGVVMFFIYAAVGLGYWWLMGIMGWLGAVVWGVVIWVGIIMACAYLLMAMWWLLAWYWVGVWLACIVCAILIIIW',
                RANDOM_HYDROPHIL : 'ERKERRQSHEQHSERRSNPRHQEHTTPERSQSTNPSPSNTKPRSPHRNSTPRRHRTHKRQQHSRKTHESDPQHQENRPPQHTQEPPSTDHPETRNRTTKNSNKHEQPTQESTSKSSPQPPKNNRQDNRQRSHQNNKNRRENDHHRETNEDPRTERHTRSHSTSKKDTNEDNESHTNNQHHSETRDHHSKRPHNNDDDKETSTDTRRDRNPDRPPKPKSDEEPKQPKRQ',
                GFP : 'MSKGEELFTGVVPILVELDGDVNGHKFSVSGEGEGDATYGKLTLKFICTTGKLPVPWPTLVTTFSYGVQCFSRYPDHMKQHDFFKSAMPEGYVQERTIFFKDDGNYKTRAEVKFEGDTLVNRIELKGIDFKEDGNILGHKLEYNYNSHNVYIMADKQKNGIKVNFKIRHNIEDGSVQLADHYQQNTPIGDGPVLLPDNHYLSTQSALSKDPNEKRDHMVLLEFVTAAGITHGMDELYK',
            }
    molarUnitMass = {'A' : 89.1, 'R' : 174.2, 'N' : 132.1, 'D' : 133.1, 'C' : 121.2,
                     'Q' : 146.2, 'E' : 147.1, 'G' : 75.1, 'H' : 155.2, 'I' : 131.2,
                     'L' : 131.2, 'K' : 146.2, 'M': 149.2, 'F' : 165.2, 'P':115.1,
                     'S' : 105.1, 'T' : 119.1, 'W' : 204.2, 'Y' : 181.2, 'V' : 117.1,
                     }
    
    def getMolarMass(gfp):
        if gfp in TDP._molarMasses:
            return TDP._molarMasses[gfp]
        
        print("Warning: No Molar mass for "+gfp+" given. Using inprecise Method to calculate. It might be adviseable to calculate it externally and add to TDP class.")
        mass = 0
        for i in TDP.sequences[gfp]:
            mass += TDP.molarUnitMass[i]
        return mass

reactionVolume = 100*10**-6 #L
mass_Cas13a = 2*10**-6 #g
molarMass_Cas13a = 140000 #g/mol



aminoAcidLength = 0.8*10**-9
avgAminoAcidVolume = 128.59*10**-30 #http://www.indiana.edu/~clemmer/Publications/pubs/pub%20051.pdf

def getNumberOfTDPs(TDPname, concentration):
    '''
    Calculate the number of TDPs on one Cas13a depending on the concentration.
    '''
    #get number of Cas13a molecules
    N_Cas13a = (mass_Cas13a / molarMass_Cas13a)*spc.N_A
    #get number of TDP molecules
    N_TDP = ((reactionVolume*concentration) / TDP.getMolarMass(TDPname))*spc.N_A
    #calculate to one Cas13a molecule
    n = N_TDP/N_Cas13a
    
    n = int(round(n))
        
    return n

def getLatticeSize(volumeFraction, binsize):
    #get reaction volume in m^3
    volume = reactionVolume*10**-3
    #get volume per Cas13a
    N_Cas13a = (mass_Cas13a / molarMass_Cas13a)*spc.N_A
    volume_Cas13a = volume/N_Cas13a
    #get volume of unit cell
    volume_unitCell = avgAminoAcidVolume * binsize
    
    #get number of unit cells necessary and dimensions of lattice depending on volume fraction
    N_unitCells = (volume_Cas13a/volume_unitCell)*volumeFraction
    cubeLatticeDimension = int(round(N_unitCells ** (1 / 3)))
    
    return cubeLatticeDimension

    
def generate3DEnvironment(TDPname, concentration, volumeFraction, binsize, beta, folder):
    '''
    generates the right number of TDPs for one Cas13a and Cas13a
    '''
    
    n = getNumberOfTDPs(TDPname=TDPname, concentration=concentration)
    
    #get lattice size for cubic lattice.
    length = getLatticeSize(volumeFraction, binsize)
    size = [length, length, length]
        
    s = Simulator(size=size, binsize=binsize, beta=beta,
                  concentration=concentration, folder=folder)

    #generate Cas13a
    CasPos = (length/2-5, length/2-5, length/2-5)
    s.f.makeProtein(CasPos, 'Units', 'Hydrophobicity', 'Charges')
    
    #get properties of the TDP
    props = s.f.getProteinAverage(TDP.sequences[TDPname])
    for i in range(n+1):
        #TODO change linktypes to not random ones.
        lts = np.random.randint(2, size=len(props)-1)
        s.f.makeChain(len(props)-1, props, lts)
        
    return s