#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Oct 29 11:44:43 2017
"""

from Generator import generate3DEnvironment

import csv
import multiprocessing
import numpy as np
import matplotlib.pyplot as plt
import itertools

class SimulationThread(multiprocessing.Process):
    callbacks = []
    folder = ""
    runParams = None
    iterations = 0
    sim = None
    
    def __init__(self, runParams, folder):
        super(SimulationThread, self).__init__()
        self.exit = multiprocessing.Event()
        
        self.runParams = runParams
        self.folder = folder
    
    def run(self):
        runP = self.runParams
        self.sim = generate3DEnvironment(TDPname=runP['TDP'], concentration=runP['conc'],
                                   volumeFraction=runP['vol'], binsize=5,
                                   beta=runP['beta'], folder=self.folder)
        
        itRange = range(self.runParams['iterations'])
        self.E_values = np.zeros_like(itRange)-1
        self.num_values = np.zeros_like(itRange)-1
        self.E_error = {}
        self.num_error = {}
        
        #determine number of bins for histogram for cube
        bins = int(self.sim.f.uk.size[0]/2)
        self.dists = {}
        
        self.numAttached = {}
        self.arates = np.zeros_like(itRange)-1
        
        CasPos = np.array(self.sim.f.uk.getPos(self.sim.f.uk.chains[0][0]))
        edges = set(itertools.combinations([0,self.sim.f.uk.size[0]]*self.sim.f.uk.dim, self.sim.f.uk.dim))
        edgedists = [np.sqrt(np.sum((np.array(edge)-CasPos)**2)) for edge in edges]
        maxDist = np.max(edgedists)
        
        
        for i in itRange:
            if self.exit.is_set():
                print("Simulation Thread Interrupted at %.f/%.f iterations, saving..."%(i, self.iterations))
                break
            
            trials = self.sim.metropolisIteration()
            energy = self.sim.energy
            num = self.sim.numberOfBindings
            
            self.arates[i] = trials
            
            #Recalculate energy at interval
            if i%10000 == 0:
                E, N = self.sim.f.getFreeEnergy()
                self.E_error[i] = (E-energy)
                energy = E
                self.sim.energy = E
                self.num_error[i] = (N-num)
                num = N
                self.sim.numberOfBindings = E
                
                
            if i%1000 == 0:
                d = self.sim.f.getProteinDistanceFromCas13a()
                #make histogram
                self.dists[i] = np.histogram(d, bins, range=(1,maxDist))
                
                self.numAttached[i] = self.sim.f.getNeighboursCas13a()
                
            
            #draw at interval
            if i%self.runParams['drawingInterval'] == 0:
                self.sim.draw(save=True, show=False)
            
            self.E_values[i] = energy
            self.num_values[i] = num
        
        #Save results
        self.sim.draw(save=True, show=False)
        
        
        self.plotAndSave(filename="energy", data=self.E_values[:i])
        self.plotAndSave(filename="energy_error", data=self.E_error)
        
        self.plotAndSave(filename="numberOfBindings", data=self.num_values[:i])
        self.plotAndSave(filename="numberOfBindings_error", data=self.num_error)
        
        self.plotAndSave(filename="NumberOfTrials", data=self.arates[:i])
        
        self.plotAndSave(filename="CAS13_attached", data=self.numAttached)
        
        #self.plotAndSave(filename="CAS13_distances", keys=self.dists.keys(), values=self.dist
        if len(self.dists) > 0:
            f = self.folder+'/CAS13_distances'
            iterations = list(self.dists.keys())
            bins = self.dists[list(self.dists.keys())[0]][1]
            with open(f+'.csv', 'w', newline='') as csvfile:
                writer = csv.writer(csvfile, delimiter=';')
                writer.writerow([0, 0, *tuple(iterations)])
                for binIndex in range(len(bins)-1):
                    values = np.zeros_like(iterations)
                    for iterationIndex, iteration in enumerate(iterations):
                        values[iterationIndex] = self.dists[iteration][0][binIndex]
                    writer.writerow([bins[binIndex], bins[binIndex+1], *tuple(values)])
        
        #Notify all listeners, that we are done!
        self._notifyAll()
    
    def plotAndSave(self, filename, data):
        if len(data) > 0:
            if type(data) is dict:
                keys = list(data.keys())
                values = list(data.values())
            else:
                keys = range(len(data))
                values = data
            
            f = self.folder+'/'+filename
            with open(f+'.csv', 'w', newline='') as csvfile:
                writer = csv.writer(csvfile, delimiter=';')
                for k in keys:
                    writer.writerow([k, data[k]])
            
            plt.ioff()
            plt.figure()
            plt.plot(list(keys), values)
            
            plt.savefig(f+'.png', transparent=True, dpi=300)
            plt.close()
    
    def addListener(self, cb):
        if not cb in self.callbacks:
            self.callbacks.append(cb)
            
    def removeListener(self, cb):
        if cb in self.callbacks:
            del(self.callbacks[self.callbacks.index(cb)])
            
    def _notifyAll(self):
        for cb in self.callbacks:
            cb()
            
    def terminate(self):
        print("Sim Thread exit init")
        self.exit.set()
