#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Sep 25 14:13:09 2017

@author: IsabellTrinh
"""
from Flask import Flask as f
from random import choice, randrange

import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator
import numpy as np
from mpl_toolkits.mplot3d import Axes3D #is required!
import logging


class Simulator:
    beta = None #1/kbT
    concentration = None #g/L
    folder = None #Where to save results?
    iterations = 0
    energy = None
    numberOfBindings = None #Number of chemical bounds
    #colorcycle = ["red", "blue", "blue", "green", "yellow", "purple", "pink", "gray", "orange"]
    colorcycle = ["C0", "C1", "C2", "C3", "C4", "C5", "C6", "C7", "C8", "C9"]
    
    def __init__(self, size, binsize, beta, concentration, folder):
        #initialize Unitkeeper in Flask
        
        #remove trailing /
        if folder[-1] == "/":
            folder = folder[:-1]
        self.folder =  folder
        
        logging.basicConfig(level=logging.DEBUG, 
                            format='[%(asctime)s] %(message)s',
                            filename=folder+"/debug.log")
        #TODO:
        #self.logH = None
        
        self.f = f(size, binsize, logging=logging)
        self.beta = beta
        self.concentration = concentration
        
                        
    def sim_snake(self):
        '''
        Simulates a snake movement in the lattice and returns the new position 
        of the idents without moving them.
        '''
        #get a random head or tail of a chain.
        head = choice(self.f.uk.chains[1:])[randrange(-1, 1)]
        
        #get open lattice sites the head can move to
        ols = self.f.uk.getOpenLatticeSites(self.f.uk.getPos(head))
        
        #randomly choose a position the head can move to if there are any.
        if ols:
            headPos = choice(ols)
            #generate new positions
            snake = self.f.snake(head, headPos)
            #logging.debug("Snake move proposed.")
            return snake
        else:
            return None
    
    
    def sim_flexibleMove(self):
        '''
        Simulates a flexible movement of a single unit and returns the new position of the idents without moving them.
        '''
        
        #get a random unit excluding the stationary Cas13a
        ident = randrange(self.f.uk.chains[1][0], len(self.f.uk.ids))
        
        #generate flexible movement
        flexible = self.f.flexibleMove(ident)
        
        #return the unit ident, the initial unit position and a random position
        # for the unit to move to 
        if flexible != None:
            logging.debug("Flexible move proposed.")
            return flexible[0], flexible[1], [choice(flexible[2])]
        else:
            return None
    
    def sim_rotate(self):
        '''
        Simulates the rotation of either the end or middle of a chain. 
        '''
        #TODO: add sim_rotateMiddle
        move_list = (self.sim_rotateEnd)
        #logging.debug("Rotation proposed.")
        return choice(move_list)()
    
    
    def sim_rotateEnd(self):
        '''
        Simulates rotation of the end of a chain (random amount of units) and returns the new position of the idents without moving them.
        '''
        #pick a random chain
        chain = choice(self.f.uk.chains[1:])
        
        #get a random head or tail of a chain as the end of the rotating part.
        endIdent = chain[randrange(-1, 1)]
        
        #get a random origin for the rotation
        origin = choice(chain)
        
        #get a random angle

        angle_list = (90, 180, 270)
        angle = choice(angle_list)
        
        rotate = self.f._rotateEnd(origin, endIdent, angle)
        
        
        if rotate != None:
            #logging.debug("End rotation proposed.")
            return rotate
        else:
            return None
        
    
    def sim_translate(self):
        '''
        Simulates translation of the chain and returns the new position of its idents without moving them.
        '''
        #pick random chain
        chain = choice(self.f.uk.chains[1:])
        chainIdent = self.f.uk.getChainIdent(chain[0])
        
        #generate a random shift
        shift = ()

        for i1 in range(0, self.f.uk.dim): 
            shift = shift + (randrange(-self.f.uk.boundaries[i1, 1], self.f.uk.boundaries[i1, 1]),)
        
        translate = self.f.translateChain(chainIdent, shift)
        
        if translate != None:
            logging.debug("Translation proposed.")
            return translate
        else:
            return None
        
        
    
    def proposeRandomMove(self):
        '''
        Randomly generates a move and returns the old and new positions of the idents.
        '''
        move = None
        while move == None:
            move_list = [self.sim_snake, self.sim_flexibleMove, self.sim_rotateEnd, self.sim_translate]
            move = choice(move_list)()
        return move
    
    
    # Monte Carlo Metropolis Method
    #
    def metropolisIteration(self):
        '''
        Simulates movement of the chains according to metropolis method. 
        Drawing interval gives the 
        '''
        
        if self.energy is None and self.numberOfBindings is None:    
            self.energy, self.numberOfBindings = self.f.getFreeEnergy()
            
        #logging.debug("Running iteration %.f"%(self.iterations+1))
        
        acc = False
        tries = 0
        while not(acc):
            tries += 1
            #generate random move and return old and new positions
            new_move = self.proposeRandomMove()
            
            #calculate energy difference between new and old positions
            E_diff, N_diff = self.f.freeEnergyDiff(ids = new_move[0], oldposs=new_move[1], newposs=new_move[2])
            #logging.debug("Energy diff: %.3f, Bind diff: %.3f"%(E_diff, N_diff))
            
            A_mov = np.exp( -self.beta * E_diff ) #Probability that change will be accepted
            ac1 = E_diff <= 0 # Downhill move always accepted
            ac2 = (E_diff > 0 and A_mov >= np.random.rand()) # Uphill move accepted with luck
            acc = ac1 or ac2
            
            #if not(acc):
            #    logging.debug("Movement rejected.")
                
    
        #logging.debug("%.f Movements rejected."%(tries-1))
        #logging.debug("Energy diff: %.3f, Bind diff: %.3f"%(E_diff, N_diff))
        #logging.debug("Movement accepted.")
        
        self.f.move(ids=new_move[0], newPoss=new_move[2])
        
        #calculate energy by subtracting E_diff
        self.energy += E_diff
        self.numberOfBindings += N_diff
        
        #logging.debug("Movement executed.")
        
        #Done
        self.iterations += 1

        #self.logH.flush()
        #self.logH.close()    
    
        return tries
    
    
    '''
    VISUALIZATION
    '''

    def draw(self, save=True, show=False):
        logging.debug("Plotting.")
        if self.f.uk.dim == 1:
            raise NotImplementedError("1D plotting not supported.")
            return None
        elif self.f.uk.dim == 2:
            raise NotImplementedError("2D plotting not supported.")
        elif self.f.uk.dim == 3:
            return self._draw3D(save=True, show=False)
        else:
            raise NotImplementedError("Ufff, more than 3D? How am I supposed to do that?")
            return None
        
    def _draw2D(self, save=True, show=False):
        #scatterplot
        
        #check dimension
        if self.f.uk.dim == 2:
            
            fig = plt.figure()
            
            #set boundaries for plot
            plt.xlim(-1, self.f.uk.boundaries[1][1]+1)
            plt.ylim(self.f.uk.boundaries[0][1]+1, -1)
            
            #get all positions of water/empty sites and plot as scatterplot
            a, b = np.argwhere(self.uk.poss == -1).T
            plt.scatter(b, a, facecolors='none', edgecolors='grey')
            
            #plot chains: for all chains in chain list
            for chain in self.f.uk.chains:
                chainIdent = self.f.uk.chains.index(chain)
                if self.f.uk.isInPeriodicBoundary(chainIdent):
                    sections = self.f.uk.getUnitsInPeriodicBoundary(chainIdent)
                    [self._plotChain(section, chainIdent) for section in sections]
                else:
                    self._plotChain(chain, chainIdent)                
            return fig
        else:
            raise RuntimeWarning("Error_draw2D: Given lattice is not 2 dimensional.")
   
         
    def _draw3D(self, save=True, show=False):
        #scatterplot
        if self.f.uk.dim == 3:
            if show:
                plt.ion()
            else:
                plt.ioff()
            
            plt.title("TDPs and CAS13 after %.f iterations"%self.iterations)
            
            fig = plt.figure()
            #set aspect ratio to equal
            #ax.set_aspect('equal')
            zoom = 2
            w, h = fig.get_size_inches()
            fig.set_size_inches(w * zoom, h * zoom)
            
            ax = fig.add_subplot(111, projection='3d')
            
            #set limits for x, y, z axes
            ax.set_xlim(self.f.uk.boundaries[0])
            ax.set_ylim(self.f.uk.boundaries[1])
            ax.set_zlim(self.f.uk.boundaries[2])

            #plot Cas13a
            chain = self.f.uk.chains[0]
            chainIdent = self.f.uk.chains.index(chain)
            sections = self.f.uk.getUnitsInPeriodicBoundary(chainIdent)
            for section in sections: 
                #get start and end unitID of chain (part)
                start = section[0]
                end = section[-1]
            
                #get position of IDs in chain and plot
                x = self.f.uk.ids[start:end+1,0]
                y = self.f.uk.ids[start:end+1,1]
                z = self.f.uk.ids[start:end+1,2]
                plt.plot(x, z, y, '-o', color = 'lightblue')
            
            
            #plot the chains
            for chain in self.f.uk.chains[1:]:
                chainIdent = self.f.uk.chains.index(chain)
                if self.f.uk.isInPeriodicBoundary(chainIdent):
                    sections = self.f.uk.getUnitsInPeriodicBoundary(chainIdent)
                    [self._plotChain(section, chainIdent) for section in sections]
                else:
                    self._plotChain(chain, chainIdent)
            
            #set label ticks to integer
            ax.xaxis.set_major_locator(MaxNLocator(integer=True))
            ax.yaxis.set_major_locator(MaxNLocator(integer=True))
            ax.zaxis.set_major_locator(MaxNLocator(integer=True))
            
            #set labels for axis
            ax.set_xlabel('x')
            ax.set_ylabel('y')
            ax.set_zlabel('z')
            
            if show:
                plt.show()
            if save:
                filename = "fullPlot_it-%.f.png"%self.iterations
                fig.savefig(self.folder +"/"+ filename, transparent=True, dpi=300)
            
            plt.close(fig)
            
        else:
            raise RuntimeWarning("Error_draw3D: Given lattice is not 3 dimensional.")
            
    def _plotChain(self, idents, chainIdent):
        #get start and end unitID of chain (part)
        start = idents[0]
        end = idents[-1]
        color = self.colorcycle[chainIdent%len(self.colorcycle)]
        
        #get position of IDs in chain and plot
        x = self.f.uk.ids[start:end+1,0]
        y = self.f.uk.ids[start:end+1,1]
        if self.f.uk.dim == 2:
            plt.plot(y, x, '-o', markersize=0.3, color=color)
            
        elif self.f.uk.dim == 3:
            z = self.f.uk.ids[start:end+1,2]
            plt.plot(x, z, y, '-o', markersize=0.3, color=color)
        #return fig