#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Aug  8 14:16:37 2017

@author: IsabellTrinh
"""

import numpy as np

class UnitKeeper:
    ids = None #array giving all unit idents with corresponding positions
    poss = None #array giving the available positions (the lattice) with corresponding unit idents
    links = None #array with link information --> first column: ident of unit, second column: chainIdent, third column: linktype (0: flexible, 1: rigid)
    dim = 0 #variable giving the chosen dimension of the lattice
    boundaries = None #array giving the valid boundaries within the array, default to all positions in the poss array
    size = None
    
    def __init__(self, size):
        #dimension equals the size of the given size array upon initiation
        self.dim = len(size)
        self.size = size
        
        #boundaries are initialized as an empty array with two columns. Set boundaries by going through size array.
        self.boundaries = np.zeros([0,2], dtype = int)
        for i in range(0, self.dim):
            self.boundaries = np.append(self.boundaries, [[0, size[i]-1]], 0)
        
        #initialize all variable arrays with right size and type.
        self.ids = np.zeros([0, self.dim], dtype = int)
        self.poss = np.zeros(size, dtype = int)-1
        self.links = np.zeros([0,2], dtype = int)
        self.props = np.zeros([0,2], dtype = int)
        self.chains = []       
        
        
       
        
    '''
    UNIT MANIPULATION
    '''
    def _addUnit(self, pos, props):
        '''
        Adds a unit at a certain position (pos) with certain properties (prop) given by an integer.
        '''
        pos = tuple(pos)
        if self.isValidPos(pos):
            if not self.isOccupied(pos):
                self.ids = np.concatenate((self.ids, [pos]))
                self.props = np.append(self.props, [props], 0)
                ident = len(self.ids)-1
                self.links = np.append(self.links, [[-1, -1]], 0)
                self.poss[pos] = ident
                return ident
            else:
                print("Error_addUnit: Site is already occupied.")
        else:
            print("Error_addUnit: pos should be ",self.dim," dimensional and is ",len(pos)," dimensional or is not within bounds.")
            return -1
               
    def moveUnit(self, ident, newPos):
        nPos = tuple(newPos)
        oldPos = self.getPos(ident)
        self.poss[oldPos] = -1
        if self.poss[nPos] != -1:
            print('WARNING: OVERWRITING POSITION')
        self.poss[nPos] = ident
        self.ids[ident] = nPos
 

    '''
    GET-FUNCTIONS FOR UNIT MANIPULATION
    '''
    def getPos(self, ident):
        if self.isValidIdent(ident):
            pos = self.ids[ident]
            return tuple(pos)
        else:
            print("Error_getPos: ident does not exist.")
        
    def getIdent(self, pos):
        pos = tuple(pos)
        if self.isValidPos(pos):
            ident = self.poss[pos]
            return ident
        else:
            print("Error_getIdent: pos should be ",self.dim," dimensional and is ",len(pos)," dimensional or is not within bounds.")

    def getProp(self, ident):
        if self.isValidIdent(ident):
            return self.props[ident]
        else: print("Error_getProp: ident does not exist")
            
    
    def getAllIdents(self, chainIdent):
        '''
        Gives all idents of either the desired chain or all idents introduced in the system
        '''
        if chainIdent == -1:
            return list(range(0, len(self.ids)))
        else:
            return self.chains[chainIdent]   
            
    
    
    
    '''
    CHAIN GENERATION AND MANIPULATION
    '''
    def addChain(self, poss, props, lts):
        chainIdent = len(self.chains)
        
        self._addUnit(poss[0], props[0])
        self.chains.append((self.getIdent(poss[0]),))
                
        for i in range(1, len(poss)):
            self._addUnit(poss[i], props[i])            
            ident = self.getIdent(poss[i])
                
            self.linkUnits(ident, chainIdent, lts[i-1])
            self.chains[chainIdent] = self.chains[chainIdent] + (ident,)
            
    def linkUnits(self, ident, chainIdent, linktype):
        '''
        Connects the chosen unit (ident) with the preceding unit with either a rigid (1) or flexible link (0).
        '''
        if self.isValidIdent(ident): 
            self.links[ident,0] = chainIdent
            self.links[ident,1] = linktype
            self.links[ident-1,0] = chainIdent
        else:
            print("Error_linkUnits: ident does not exist or the previous unit (ident-1) is not on a neighbouring site.")
    
    def moveChainPart(self, ids, newPoss):
        '''
        Moves a part of a chain or the whole chain to new positions.
        '''

        #Get units of the chain and their positions
        oldPoss = [self.getPos(x) for x in ids]
            
        #delete all units in the lattice
        for pos in oldPoss:    
            self.poss[pos] = -1
        
        #change positions of the units in the ident list
        for unit in ids:
            self.ids[unit] = newPoss[unit-ids[0]]
        
        #put unit idents into the lattice
        for pos in newPoss:
            self.poss[pos] = ids[newPoss.index(pos)]
            
    def addProtein(self, poss, props):
        chainIdent = len(self.chains)
        
        self._addUnit(poss[0], props[0])
        self.chains.append((self.getIdent(poss[0]),))
        
        for i in range(1, len(poss)):
            self._addUnit(poss[i], props[i])
            ident = self.getIdent(poss[i])
            
            self.chains[chainIdent] = self.chains[chainIdent] + (ident,)
            self.links[ident,0] = chainIdent
            self.links[ident,1] = -2
            self.links[ident-1, 0] = chainIdent
    
    def applyPeriodicBoundary(self, pos):
        '''
        Applies the periodic boundaries to a given position and returns pos in periodic boundary
        '''
        mod = [n%axis for n, axis in zip(pos, self.size)]
        
        return tuple(mod)
    
    def reversePeriodicBoundary(self, chainIdent):
        #initiate list of old and new positions 
        oldPoss = []
        newPoss = []
        
        #get the sections of the chain that are between periodic boundaries.
        sections = self.getUnitsInPeriodicBoundary(chainIdent)
        
        #Add positions of first section to newPos, as it remains unchanged
        [oldPoss.append(self.getPos(unit)) for section in sections for unit in section]
        
        #copy list of old positions to change later.
        newPoss = oldPoss[:]
        
        #get which axis passed the boundary
        sub = [np.subtract(self.getPos(section[-1]), self.getPos(section[-1]+1)) for section in sections if not section == sections[-1]]
        
        #for each boundary passed (number of elements in axis), add boundary to sections[1:]
        for a in sub:
            #get index of axis and transform into number
            axis = np.where(a != 0)
            axis = axis[0][0]
            #get final arrays to add for each section, depending on sign of the axis array
            if a[axis] > 0:
                a[axis] = self.size[axis]
            else:
                a[axis] = -self.size[axis]
               
        #get new positions
        for i in range(len(sub)):
            #get sections to add to
            section = sections[i+1:]
            #flatten list to get idents
            indices = [ident-sections[0][0] for sec in section for ident in sec]
            #add/subtract the axis to/from the section
            for index in indices:
                newPoss[index] = tuple(np.add(newPoss[index], sub[i]))
        
        return oldPoss, newPoss
            
        


    '''
    GET-FUNCTIONS FOR CHAIN GENERATION
    '''
    def getNeighbours(self,pos):
        '''
        Gets all neighbours of a given position, even outside of boundaries.
        '''
        
        #get any possible neighbours for either 1D, 2D or 3D
        if self.dim == 1:
            anyNb = ([pos-1],[pos+1])
        elif self.dim == 2:
            anyNb = [(pos[0]-1, pos[1]), (pos[0], pos[1]-1), (pos[0]+1, pos[1]), (pos[0], pos[1]+1)]
        elif self.dim == 3:
            anyNb = [(pos[0]-1, pos[1], pos[2]), (pos[0], pos[1]-1, pos[2]), (pos[0], pos[1], pos[2]-1), (pos[0]+1, pos[1], pos[2]), (pos[0], pos[1]+1, pos[2]), (pos[0], pos[1], pos[2]+1)]
        else:
            print("Error!")
            
        return anyNb
    
    def getNeighboursInBoundary(self, pos):
        '''
        Gets all neighbours of a given position and applies periodic boundary conditions.
        '''
        #initiate neighbour list in boundary
        neighbours = []
        anyNb = self.getNeighbours(pos)
        
        #apply periodic boundary conditions
        neighbours = [self.applyPeriodicBoundary(neighbour) for neighbour in anyNb]
        
        #double check if all positions are within boundary
        realNb = [n for n in neighbours if self.isValidPos(n)]
        
        return realNb
    
    def getOpenLatticeSites(self, pos):
        '''
        Returns an array with the positions of unoccupied lattice sites one row away from the unit in question.
        '''
        nbs = self.getNeighboursInBoundary(pos)
        ols = [n for n in nbs if not self.isOccupied(n)]
        return ols 
    
    def getNumberOfNeighbours(self, ident):
        '''
        Returns the number of neighbours on the polymer chain.
        '''
        if self.isValidIdent(ident):
            if self.isChainEnd(ident):
                return 1
            else:
                return 2    
    
    
    
    
    '''
    GET-FUNCTIONS FOR CHAINS
    '''
    def getChainIdent(self, ident):
        if self.isValidIdent(ident):
            return self.links[ident][0]
        else:
            print("Error_getChainIdent: Ident does not exist.")
            
    def _getLinktype(self, ident1, ident2):
        '''
        Returns '1' if link is rigid and '0' if link is flexible, '-1' if they are not linked.
        NOT VALID
        '''
        if self.isValidIdent(ident1) and self.isValidIdent(ident2):
            
        # check if both idents are on neighbouring sites and are behind each other in the same chain
            if self.areNeighbours(ident1, ident2) and abs(ident1-ident2) == 1 and self.getChainIdent(ident1) == self.getChainIdent(ident2): 
                #the higher ident gives the linktype information
                return self.links[max(ident1, ident2),1]
            else: 
                print("Error_getLinktype: Unit ",ident1," and unit ",ident2," are not linked.")
        else:
            print("Error_getLinktype: idents do not exist.")
        
    def getChainNeighbours(self, ident):
        '''
        Returns the IDs of the neighbours of the given ident on the chain
        '''
        #check if ident is the start of a chain
        chainIdent = self.getChainIdent(ident)
        if ident == self.chains[chainIdent][0]:
            return [ident+1]
        #check if ident is the end of a chain  
        elif ident == self.chains[chainIdent][-1]:
            return [ident-1]
        #otherwise the ident is in the middle of the chain
        else:
            return [ident-1, ident+1]
    
    def getRigidNeighbours(self, ident):
        '''
        Returns the ID of chain neighbours connected by a rigid link (e.g. part of an alpha-helix)
        '''
        rn = []
        if self.isValidIdent(ident):
            cn = self.getChainNeighbours(ident)
            for n in cn:
                if self._getLinktype(n, ident) == 1:
                    rn.append(n)
            return rn
        
    def getUnitsInPeriodicBoundary(self, chainIdent):
        '''
        Checks if the chain is on the border of the periodic boundary and has 
        to be plotted separately.
        '''
        #initiate chain and sections
        chain = self.chains[chainIdent]
        sections = []
        if self.isValidChainIdent(chainIdent) and self.isInPeriodicBoundary(chainIdent):
            #check if consecutive elements in chain are neighbours
            l = [self.areNeighbours(unit, neighbour) for unit, neighbour in zip(chain, chain[1:])]
            idents = [i for i, x in enumerate(l) if not x]
            
            #split the first section of the chain and append to sections.
            sections.append(chain[0:idents[0]+1])
            
            #split the middle parts of the chain in sections for boundary
            [sections.append(chain[start+1:end+1]) for start, end in zip(idents, idents[1:])]
            
            #append the remainder of the chain as section
            sections.append(chain[idents[-1]+1:chain[-1]+1])
            
            return sections    
        
        else:
            return chain
            print("Error_getUnitsInPeriodicBoundary: ChainID is not valid or chain is not on periodic boundary")
    


        
    '''
    STATUS CHECK FUNCTIONS
    '''  

    '''CHAINS'''    
    def areNeighbours(self, ident1, ident2):
        #TODO: consider periodic boundaries 
        '''
        Checks if ident2 is on a neighbouring site in the lattice to ident1 within periodic 
        boundaries. Neighbours crossing boundaries are not taken into account.
        '''
        if self.isValidIdent(ident1) and self.isValidIdent(ident2):
            pos = self.getPos(ident1)
            l = self.getNeighbours(pos)
            return any(np.equal(l, self.getPos(ident2)).all(1))
        else:
            print("Error_areNeighbours: One or both of the given idents do not exist.")   
            
    def isInPeriodicBoundary(self, chainIdent):
        '''
        Checks if the chain is on the border of the periodic boundary and has 
        to be plotted separately.
        '''
        chain = self.chains[chainIdent]
        if self.isValidChainIdent(chainIdent):
            #check if consecutive elements in chain are neighbours
            a = [self.areNeighbours(unit, neighbour) for unit, neighbour in zip(chain, chain[1:])]
            
            #return false if all consecutive elements in chain are neighbours
            #return true if not all consecutive elements in chain are neighbours 
            return not all(a)
        
        else:
            print("Error_inPeriodicBoundary: ChainIdent is not valid.")

    def isInChain(self, ident, chainIdent):
        '''
        Checks if the given unit is in the given chain
        '''
        if self.isValidIdent(ident):
            return self.links[ident,0] == chainIdent
    
    def isChainEnd(self, ident):
        '''
        Checks if the unit is on the end of a chain.
        '''
        chainIdent = self.getChainIdent(ident)
        return ident == self.chains[chainIdent][-1] or ident == self.chains[chainIdent][0]
 
    def isValidChainIdent(self, chainIdent):
        '''
        Checks if the given chain exists.
        '''
        return chainIdent < len(self.chains)
    
    '''UNITS'''
    def isOccupied(self, pos):
        '''
        Returns 'true' if position is already occupied and 'false' if it is empty.
        '''
        if self.isValidPos(pos):
            return self.getIdent(pos) != -1 
        
    def isValidIdent(self, ident):
        '''
        Checks if ident is already introduced.
        '''
        return ident < len(self.ids) and ident > -1
        
    def isValidPos(self, pos):
        '''
        Checks if pos is the right dimensionality and within bounds.
        '''
        return self._isWithinBounds(pos) and self._hasRightDimension(pos)
    
    def _hasRightDimension(self, pos):
        '''
        Checks if the dimension of the given position matches the overall dimension of the lattice.
        '''
        return len(pos) == self.dim
    
    def _isWithinBounds(self, pos):
        '''
        Checks if the given position is within the boundaries set for the lattice. changeBoundaries changes these.
        '''
        return all([a >= b for a, b in zip((pos), self.boundaries[::1,0])]) and all([a <= b for a, b in zip((pos), self.boundaries[::1,1])])